import copy
import warnings
import weakref

import numpy as np

import marquetry


# ==================================================
# ContainerNode / Container
# ==================================================
try:
    import cupy
    allow_array = (np.ndarray, cupy.ndarray)
except ImportError:
    cupy = np
    allow_array = (np.ndarray,)


class ContainerNode(object):
    """Node in the backward computational graph representing arrays.

        This object represents a container node in a computational graph.
        The node is used in backward computation to determine which gradient to be passed to each function.

        A container node is held by the corresponding :class:`Container` object,
        which is defined by user. :class:`Function` objects that take the container
        as an input also hold references to the container node.

        Note that the node does not hold a reference to the corresponding data
        array in general.
        When the function call :meth:`retain_input` and/or :meth:`retain_output`,
        node store the corresponding array.

        Users usually do not need to touch this container node object.
        The computational graph is automatically managed by Marquetry,
        and any interface that is beneficial for users is also provided by :class:`Container`.

        Args:
            container (Container): The corresponding container object.
            name (str): Name of the container node.

    """

    def __init__(self, container, name):
        self._container = weakref.ref(container)
        self._creator = None
        self._data = None
        self._generation = 0
        self._name = name
        self._grad = None
        self._ndim = None
        self._dtype = None
        self._shape = None

    @property
    def creator(self):
        """creator name which create this container."""
        return self._creator

    @creator.setter
    def creator(self, func):
        self._creator = func
        if func is not None:
            self._generation = func.generation + 1

    @property
    def data(self):
        """Data array corresponding container.

        If the data is already released, this return `None`.
        """
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self.set_data_type(d)

    @property
    def grad(self):
        """Gradient array corresponding container."""
        return self._grad

    @grad.setter
    def grad(self, g):
        self._grad = g

    @property
    def label(self):
        """Short text explains this container node."""
        if self.shape == ():
            return str(self.dtype)

        return "(%s), %s" % (", ".join(map(str, self.shape)), str(self.dtype))

    @property
    def name(self):
        """Get the name assigned to this corresponding Container."""
        return self._name

    @property
    def generation(self):
        """Generation number of the corresponding container, which is used for backpropagation computation."""
        return self._generation

    @property
    def ndim(self):
        """Number of dimensions of the data array."""
        return self._ndim

    @ndim.setter
    def ndim(self, data_ndim):
        self._ndim = data_ndim

    @property
    def dtype(self):
        """Dtype of the data array."""
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        self._dtype = dtype

    @property
    def shape(self):
        """Shape of the data array."""
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape

    def set_creator(self, creator):
        """Set the creator function for this container node."""
        self.creator = creator

    def unchain(self):
        """Remove the creator function reference."""
        self.creator = None

    def retain_data(self):
        """Retain container data."""
        container = self._container()
        if container is not None:
            self.data = container.data
        else:
            raise RuntimeError("Cannot retain container data: the container has been already released.")

    def set_data_type(self, d):
        """Set data type and shape attributes based on the data array."""
        if d is None:
            self.dtype = None
            self.shape = None
            self.ndim = None
        else:
            self.dtype = d.dtype
            self.shape = d.shape
            self.ndim = d.ndim

    def set_grad_with_check(self, g, func, data):
        """Set the gradient while checking its type."""
        _check_grad_type(func, data, g)
        self._grad = g


def _check_grad_type(func, x, grad_x):
    """Check the type, dtype, and shape compatibility between data and gradient.

        This function is used to ensure that the type, dtype, and shape of the data and its gradient match.
        It checks if the gradient data has the same type as the input data, has the same dtype, and has the same shape.

        Args:
            func (Function or None): The function that generated the gradient, if available.
            x (Container): The input container.
            grad_x (Container): The gradient container.

        Raises:
            TypeError: If the data type or dtype of the gradient does not match the input data,
                or if the shapes do not match.

        Note:
            If either the input data or the gradient data is `None`, no check is performed.
    """

    def make_message(message):
        if func:
            detail = "Function `{0}` ({1}) has a bug.\n".format(
                type(func).__name__, func.name
            )
            detail += '''
            Please report this error to developer with the issue trace.
            '''

        else:
            detail = ""

        detail += message
        return detail

    if x.data is None or grad_x is None:
        return

    if not isinstance(grad_x.data, type(x.data)):
        msg = ("Type of data and grad mismatch\n {} ≠ {}".format(type(x.data), type(grad_x.data)))

        raise TypeError(make_message(msg))

    if grad_x.dtype != x.data.dtype:
        raise TypeError("data and grad dtype mismatch.")

    if grad_x.shape != x.data.shape:
        raise ValueError("grad and data shape mismatch.")


class Container(object):
    """Represents a container with associated data and gradient in a computational graph.

        This class is used to create containers that hold data and gradients in a computational graph.
        Containers are typically　created by wrapping data arrays using this class.

        Args:
            data (:class:`numpy.ndarray` or :class:`cupy.ndarray` or None):
                The data associated with the container.
            name (str or None): A name or label for the container,
                which is helpful for debugging and visualization.

        Note:
            Containers created with this class are used in Marquetry's automatic differentiation framework.
            The `Container` class wraps data and associates it with gradient information.
            Containers can be connected through functions to create computational graphs for automatic differentiation.

    """

    __array_priority__ = 200

    def __init__(self, data, name=None):
        if data is not None and not isinstance(data, allow_array):
            raise TypeError("{} is not supported.".format(type(data)))

        self._data = data
        self._name = name
        self._node = ContainerNode(self, name)

        self._iteration = 0

    @property
    def creator(self):
        """The `Function` object that created this container during forward computation.

            If `creator` is `None`, it means that this container is not the result of a function,
            and backpropagation will not proceed beyond this container.

            Returns:
                Function or None: The function that created this container, 
                or `None` if the container is not created by a function.
        """
        return self._node.creator

    @creator.setter
    def creator(self, func):
        self._node.creator = func

    @property
    def data(self):
        """Get the data array associated with this container.

            Returns:
                ndarray or None: The data array associated with this container.
                    If no data is associated, it returns `None`.
        """
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self._node.set_data_type(d)

    @property
    def node(self):
        """Get the corresponding ContainerNode for this container.

            Returns:
                ContainerNode: The ContainerNode corresponding to this container.
        """
        return self._node

    @property
    def grad(self):
        """Get the gradient array associated with this container.

            Returns:
                Container: The gradient container corresponding to this container.
        """
        return self._node.grad

    @grad.setter
    def grad(self, g):
        self._node.set_grad_with_check(g, None, self)

    @property
    def generation(self):
        """Get the generation number associated with this container.

            Returns:
                int: The generation number used for backpropagation.
        """
        return self._node.generation

    def set_creator(self, func):
        self._node.set_creator(func)

    def unchain(self):
        """Remove the link to the creator function.

            Note:
                Calling this method removes the computational graph connection between this container
                and its creator function, effectively stopping backpropagation at this point.
        """
        self.creator = None

    def unchain_backward(self):
        """Unchain this container and its ancestors in the computational graph.

            This method unchains the current container and its ancestors by removing the links
            between them and their creator functions. It effectively stops backpropagation
            from this container onward.

            Note:
                This method iteratively unchains all ancestors of the current container,
                ensuring that backpropagation is halted at this point.
        """

        funcs = []
        seen_set = set()

        def add_func(f):
            if f not in seen_set:
                funcs.append(f)
                seen_set.add(f)

        add_func(self.creator)

        while funcs:
            f = funcs.pop()
            for x in f.inputs:
                if x.creator is not None:
                    add_func(x.creator)
                    x.unchain()

    def retain_data(self):
        """Retain the data of this container within its corresponding ContainerNode.

            When called, this method sets the data of this container to the data attribute
            of its corresponding ContainerNode. This is useful when you want to ensure that
            the data is retained within the computational graph for later use.

            Note:
                Retaining data can be helpful when you need to ensure that the data remains
                accessible within the computational graph, even if the original Container object
                goes out of scope.

        """
        self._node.data = self._data

    def clear_grad(self):
        """Clear the gradient of this container.

            When called, this method sets the gradient of this container to None,
            effectively clearing any gradient information associated with it.
            This can be useful when you want to remove gradient information
            from a container(corresponding container node).
        """

        self.grad = None

    def backward(self, retain_grad=False):
        """Perform backpropagation starting from this container.

            During backpropagation, the `backward` method of each parent function, which created
            this container, is called. This process calculates and stores the gradient for this
            container (to the corresponding container node).

            If the gradient (`grad`) of this container is initially `None`, this method initializes
            it with an array of ones with the same shape as the data.
            This serves as the initial error for the backpropagation process.

            Args:
                retain_grad (bool): If `True`, retain the gradient computation graph
                    for potential higher-order derivatives.
                    Otherwise, release computational graph of the backpropagation.

            Note:
                In most cases of model training, you don't need higher-order derivatives,
                so it is recommended to set `retain_grad` to `False`.
                This method go reverse through the computational graph created during forward computation
                and calculates gradients for each input container.
                The `backward` method is typically called after the forward pass to compute gradients
                and update model parameters.

        """
        if self.creator is None:
            return

        if self.grad is None:
            xp = marquetry.cuda_backend.get_array_module(self.data)
            self.grad = Container(xp.ones_like(self.data))

        funcs = []
        seen_set = set()

        def add_func(f):
            if f not in seen_set:
                funcs.append(f)
                seen_set.add(f)
                funcs.sort(key=lambda x: x.generation)

        add_func(self.creator)

        while funcs:
            f = funcs.pop()
            outputs = [y() for y in f.outputs]
            grad_ys = tuple(
                [None if output is None else output.grad for output in outputs])

            in_data = tuple([x.data for x in f.inputs])

            f.output_data = tuple(
                [None if y is None else y.data for y in outputs])

            grad_xs = f.backward(in_data, grad_ys)

            if not getattr(f, "_output_retain_ever", False):
                f.output_data = None

            if not isinstance(grad_xs, tuple):
                if isinstance(grad_xs, list):
                    grad_xs = tuple(grad_xs)
                else:
                    grad_xs = (grad_xs,)

            for x, grad_x in zip(f.inputs, grad_xs):
                if x.grad is None:
                    x.grad = grad_x
                else:
                    x.grad = x.grad + grad_x

                if x.creator is not None:
                    add_func(x.creator)

            if not retain_grad:
                for y in f.outputs:
                    y().grad = None

            del grad_xs

    @property
    def shape(self):
        """Get the shape of the data array.

            Returns:
                tuple: A tuple representing the shape of the data array.
                    For example, for a 2D array, the shape might be (rows, columns).

        """
        return self.data.shape

    @property
    def ndim(self):
        """Get the dimensionality of the data array.

            Returns:
                int: The number of dimensions in the data array.

        """
        return self.data.ndim

    @property
    def size(self):
        """Get the total number of elements in the data array.

            Returns:
                int: The total number of elements in the data array.

        """
        return self.data.size

    @property
    def dtype(self):
        """Get the data type of the data array.

            Returns:
                numpy.dtype: The data type of the elements in the data array.

            Note:
                The `dtype` property allows you to access the data type of the elements in the
                data array, such as float32, int64, etc.
        """
        return self.data.dtype

    @property
    def T(self):
        """Transpose the data array (reversed axis).

            Returns:
                Container: A new container containing the transposed data array.

            Examples:
                >>> x = Container(np.array([[1, 2, 3], [2, 3, 4]]))
                >>> x.shape
                (2, 3)
                >>> x.T.shape
                (3, 2)
        """
        return marquetry.functions.transpose(self)

    @property
    def name(self):
        """Get the name assigned to this Container.

            Returns:
                str or None: The name assigned to this Container, or None if no name is assigned.

            Examples:
                >>> x = Container(np.array(3), name="my_container")
                >>> x.name
                'my_container'
        """

        return self._name

    def astype(self, dtype, inplace=False):
        """Return a new container with a specified data type.

            Args:
                dtype (str or numpy.dtype): The desired data type for the new container.
                inplace (bool): If True, the data type of the current container is
                    modified in-place. If False (default), a new container with the specified
                    data type is returned.

            Returns:
                Container: A new container with the specified data type.

            Examples:
                >>> x = Container(np.array([1, 2, 3], dtype='int32'))
                >>> x.dtype
                dtype('int32')
                >>> x.astype("float64", inplace=True)
                >>> x.dtype
                dtype('float64')
                >>> y = x.astype("float32", inplace=False)
                >>> x.dtype
                dtype('float64')
                >>> y.dtype
                dtype('float32')

        """

        if inplace:
            self._data = self._data.astype(dtype)
        else:
            data = self.copy()
            data = data._data.astype(dtype)
            return data

    def copy(self):
        """Create a copy of this container.

            This method copy only data, not gradient and computational graph.

            Returns:
                Container: A new container that has a copy of the current container data array.

            Examples:
                >>> x = Container(np.array([1, 2, 3]))
                >>> y = x.copy()
                >>> y.data[0] = 99
                >>> x.data[0]
                1
        """
        data_array = self.data.copy()
        copy_output = Container(data_array, name=self.name)
        return copy_output

    def dot(self, other):
        """Calculate the dot product between this container and another container or array.

            Args:
                other (Container or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The other container to compute the dot product with.

            Returns:
                Container: A new container containing the dot product of this container and
                    the other container.

            Examples:
                >>> x = Container(np.array([1, 2, 3]))
                >>> y = Container(np.array([4, 5, 6]))
                >>> x.dot(y).data
                container(32)

        """
        return marquetry.functions.matmul(self, other)

    def max(self, axis=None, keepdims=False):
        """Calculate the maximum value along the specified axis.

            Args:
                axis (int or tuple or None): The axis or axes along which to compute the maximum value.
                    If None (default), the maximum value is calculated over all elements.
                keepdims (bool): If True, the dimensions of the output are retained,
                    and the result will have the same number of dimensions as the original container.
                    If False (default), dimensions of size 1 are removed from the output.

            Returns:
                Container: A new container containing the maximum value along the specified axis or axes.

            Examples:
                >>> x = Container(np.array([[1, 2], [3, 4]]))
                >>> x.max()
                container(4)
                >>> x.max(axis=0)
                container([3 4])
                >>> x.max(axis=1, keepdims=True)
                container([[2]
                           [4]])
        """
        return marquetry.functions.max(self, axis, keepdims)

    def mean(self, axis=None, keepdims=False):
        """Calculate the mean (average) along the specified axis.

            Args:
                axis (int or tuple or None): The axis or axes along which to compute the mean.
                    If None (default), the mean is calculated over all elements.
                keepdims (bool): If True, the dimensions of the output are retained,
                    and the result will have the same number of dimensions as the original container.
                    If False (default), dimensions of size 1 are removed from the output.

            Returns:
                Container: A new container containing the mean along the specified axis or axes.

            Examples:
                >>> x = Container(np.array([[1, 2], [3, 4]]))
                >>> x.mean()
                container(2.5)
                >>> x.mean(axis=0)
                container([2. 3.])
                >>> x.mean(axis=1, keepdims=True)
                container([[1.5]
                           [3.5]])
        """

        return marquetry.functions.mean(self, axis, keepdims)

    def min(self, axis=None, keepdims=False):
        """Calculate the minimum value along the specified axis.

            Args:
                axis (int or tuple or None): The axis or axes along which to compute the minimum value.
                    If None (default), the minimum value is calculated over all elements.
                keepdims (bool): If True, the dimensions of the output are retained,
                    and the result will have the same number of dimensions as the original container.
                    If False (default), dimensions of size 1 are removed from the output.

            Returns:
                Container: A new container containing the minimum value along the specified axis or axes.

            Examples:
                >>> x = Container(np.array([[1, 2], [3, 4]]))
                >>> x.min()
                container(1)
                >>> x.min(axis=0)
                container([1 2])
                >>> x.min(axis=1, keepdims=True)
                container([[1]
                           [3]])
        """

        return marquetry.functions.min(self, axis, keepdims)

    def repeat(self, repeats, axis=None):
        """Repeat elements of the container along the specified axis.

            Args:
                repeats (int or list of ints or tuple of ints):
                    The number of times to repeat each element along the specified axis.
                    If it is an int, all elements are repeated that number of times.
                    If it is a sequence of ints, it must have the same length as the specified axis.
                axis (int): The axis along which to repeat the elements.
                    If None (default), the container is flattened before repeating.

            Returns:
                Container: A new container containing the repeated elements along the specified
                    axis.

            Examples:
                >>> x = Container(np.array([1, 2, 3]))
                >>> x.repeat(3)
                array([1, 1, 1, 2, 2, 2, 3, 3, 3])
                >>> x.repeat([2, 3, 1])
                array([1, 1, 2, 2, 2, 3])
        """
        return marquetry.functions.repeat(self, repeats, axis)

    def reshape(self, *shape):
        """Reshape the container to the specified shape.

            Args:
                *shape: One or more integers or a tuple of integers specifying the desired
                    shape of the reshaped container.

            Returns:
                Container: A new container with the specified shape.

            Examples:
                >>> x = Container(np.array([1, 2, 3, 4, 5, 6]))
                >>> x.reshape(2, 3)
                container([[1 2 3]
                           [4 5 6]])
                >>> x.reshape(-1, 2)
                container([[1 2]
                           [3 4]
                           [5 6]])

        """

        if len(shape) == 1 and isinstance(shape[0], (tuple, list)):
            shape = shape[0]
        return marquetry.functions.reshape(self, shape)

    def sum(self, axis=None, keepdims=False):
        """Calculate the sum of the container's elements along the specified axis.

            Args:
                axis (int or tuple or None): The axis or axes along which to compute the sum.
                    If None (default), the sum is calculated over all elements.
                keepdims (bool): If True, the dimensions of the output are retained,
                    and the result will have the same number of dimensions as the original container.
                    If False (default), dimensions of size 1 are removed from the output.

            Returns:
                Container: A new container containing the sum along the specified axis or axes.

            Examples:
                >>> x = Container(np.array([[1, 2], [3, 4]]))
                >>> x.sum()
                container(10)
                >>> x.sum(axis=0)
                container([4 6])
                >>> x.sum(axis=1, keepdims=True)
                container([[3]
                           [7]])
        """

        return marquetry.functions.sum(self, axis, keepdims)

    def squeeze(self, axis=None):
        """Remove single-dimensional entries from the container's shape.

            Args:
                axis (int or tuple of ints or list of ints or None):
                    The axis along which to remove single-dimensional entries.
                    If None (default), all single-dimensional entries are removed.

            Returns:
                Container: A new container with the specified single-dimensional entries removed.

            Examples:
                >>> x = Container(np.array([[[1]], [[2]], [[3]]]))
                >>> x.shape
                (3, 1, 1)
                >>> y = x.squeeze()
                >>> y.shape
                (3,)
                >>> z = x.squeeze(axis=1)
                >>> z.shape
                (3, 1)
        """
        return marquetry.functions.squeeze(self, axis)

    def to_numpy(self):
        """Convert the container to a NumPy array.

            Returns:
                np.ndarray: A NumPy array representing the data of the container.

            Raises:
                TypeError: If the container has a gradient container, it cannot be converted to a NumPy array.

            Examples:
                >>> x = Container(np.array([1, 2, 3]))
                >>> y = x.to_numpy()
                >>> type(y)
                <class 'numpy.ndarray'>
        """

        if self.grad is not None:
            raise TypeError("The container has gradient so it can't convert to numpy array. please do clear_grad first.")
        return self.data

    def transpose(self, *axes):
        """Transpose the container's data array.

            Args:
                *axes: One or more integers or a tuple of integers specifying the desired
                    order of dimensions in the transposed container. If no axes are provided,
                    the container is fully transposed.

            Returns:
                Container: A new container with the dimensions reordered as specified.

            Examples:
                >>> x = Container(np.array([[[1], [2], [3]], [[4], [5], [6]]]))
                >>> x.shape
                (2, 3, 1)
                >>> y = x.transpose()
                >>> y.shape
                (1, 3, 2)
                >>> z = x.transpose(1, 0, 2)
                >>> z.shape
                (3, 2, 1)

        """

        if len(axes) == 0:
            axes = None
        elif len(axes) == 1:
            if isinstance(axes[0], (tuple, list)) or axes[0] is None:
                axes = axes[0]

        return marquetry.functions.transpose(self, axes)

    def unsqueeze(self, axis):
        """Add a single-dimensional entry to the container's shape.

            Args:
                axis (int): The axis along which to add a single-dimensional entry.

            Returns:
                Container: A new container with a single-dimensional entry added along the specified axis.

            Examples:
                >>> x = Container(np.array([1, 2, 3]))
                >>> y = x.unsqueeze(0)
                >>> y.shape
                (1, 3)
                >>> z = x.unsqueeze(1)
                >>> z.shape
                (3, 1)

        """

        return marquetry.functions.unsqueeze(self, axis)

    def to_cpu(self):
        """Move the container's data array to the CPU memory(numpy.ndarray).

            This method converts the container's data array from GPU memory to CPU memory.
            If the container's data is already in CPU memory or if it is `None`, no action is taken.

        """

        if self.data is None:
            return

        self._data = marquetry.cuda_backend.as_numpy(self.data)

        node = self._node
        if node.data is not None:
            node.retain_data()

    def to_gpu(self):
        """Move the container's data array to GPU memory(cupy.ndarray).

            This method converts the container's data array from CPU memory to GPU memory.
            If the container's data is already in GPU memory or if it is `None`, no action is taken.

        """

        if self.data is not None:
            self._data = marquetry.cuda_backend.as_cupy(self.data)

            node = self._node
            if node.data is not None:
                node.retain_data()

    def __matmul__(self, other):
        return marquetry.functions.matmul(self, other)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        if self.data is None:
            return "container(None)"
        p = str(self.data).replace("\n", "\n" + " " * 10)
        return "container(" + p + ")"

    def __add__(self, other):
        return marquetry.functions.add(self, other)

    def __radd__(self, other):
        return marquetry.functions.add(self, other)

    def __iadd__(self, other):
        return marquetry.functions.add(self, other)

    def __sub__(self, other):
        return marquetry.functions.sub(self, other)

    def __rsub__(self, other):
        return marquetry.functions.rsub(self, other)

    def __isub__(self, other):
        return marquetry.functions.sub(self, other)

    def __mul__(self, other):
        return marquetry.functions.mul(self, other)

    def __rmul__(self, other):
        return marquetry.functions.mul(self, other)

    def __imul__(self, other):
        return marquetry.functions.mul(self, other)

    def __neg__(self):
        return marquetry.functions.neg(self)

    def __truediv__(self, other):
        return marquetry.functions.div(self, other)

    def __rtruediv__(self, other):
        return marquetry.functions.rdiv(self, other)

    def __itruediv__(self, other):
        return marquetry.functions.div(self, other)

    def __pow__(self, power):
        return marquetry.functions.pow(self, power)

    def __getitem__(self, item):
        return marquetry.functions.get_item(self, item)

    def __eq__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data == other.data)

        return Container(array_data)

    def __ne__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data != other.data)

        return Container(array_data)

    def __lt__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data < other.data)

        return Container(array_data)

    def __gt__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data < other.data)

        return Container(array_data)

    def __le__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data <= other.data)

        return Container(array_data)

    def __ge__(self, other):
        xp = marquetry.cuda_backend.get_array_module(self)
        other = as_container(as_array(other, marquetry.cuda_backend.get_array_module(self.data)))
        array_data = xp.asarray(self.data >= other.data)

        return Container(array_data)

    def __bool__(self):
        return self.data.__bool__()

    def __hash__(self):
        return super(Container, self).__hash__()


class Parameter(Container):
    """A special type of container used to represent model parameters.

        In deep learning frameworks, the `Parameter` class is typically used to
        represent parameters like weights and biases.
        It inherits from the `Container` class, which means it can be used like a regular container
        but often carries specific attributes or methods tailored for model parameters.

        This class provides the flexibility to add parameter-specific functionality
        or attributes when needed.

        Examples:
            >>> weight = Parameter(np.random.randn(10, 5), name="weight")
            >>> bias = Parameter(np.zeros(5), name="bias")
    """
    pass


def as_container(obj):
    """Convert an object to a Container if it is not already one.

        Args:
            obj: The object to be converted to a Container.

        Returns:
            Container: The Container representation of the input object if it is not already a Container.

        Examples:
            >>> x = np.array(5)
            >>> type(x)
            <class 'numpy.ndarray'>
            >>> y = as_container(x)
            >>> type(y)
            <class 'marquetry.container.Container'>
    """

    if isinstance(obj, Container):
        return obj
    return Container(obj)


def as_array(x, array_type=np):
    """Convert an integer to a ndarray whose specified module.

        If the input object is a scalar, it is converted to a ndarray whose specified module.
        Otherwise, the object is returned as is.

        Args:
            x(int or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The object to be converted to a NumPy array if it is a scalar.
            array_type (numpy or cupy): The module managed the actual array object
                to use when converting a scalar to a ndarray, selectable cupy or numpy.
                Defaults to `np`, which is the NumPy module.

        Returns:
            :class:`np.ndarray` or :class:`cp.ndarray`:
                The ndarray representation of the input object if it is a scalar.
                Otherwise, the input object itself.

        Examples:
            >>> x = 5
            >>> y = as_array(x)
            >>> type(y)
            <class 'numpy.ndarray'>

    """

    if np.isscalar(x):
        return array_type.array(x)
    return x


def array(x, name=None):
    """Create a Container from a given ndarray whose CuPy or NumPy or list of numbers or raw int or float.

        Args:
            x(int or float or :class:`numpy.ndarray` or :class:`cupy.ndarray` or list of number):
                The ndarray to be wrapped in a Container.
            name (str or None): The name to assign to the Container.

        Returns:
            Container: A Container object that wraps the input NumPy array.

        Examples:
            >>> import numpy as np
            >>> arr = np.array([1, 2, 3])
            >>> var = array(arr, name="my_container")
            >>> type(var)
            <class 'marquetry.container.Container'>
            >>> var.name
            'my_container'
    """
    if isinstance(x, list) and all(isinstance(data, (int, float)) for data in x):
        x = np.array(x)

    elif isinstance(x, (int, float)):
        x = np.array(x)

    elif not isinstance(x, allow_array):
        raise ValueError("x for array() should be list of ints or ndarray, but got {}".format(type(x)))

    return Container(x, name=name)


def random(shape, name=None, random_std=1.):
    """Generate a random Container with a specified shape and standard deviation.

        This function creates a random array container with the specified shape and scales it
        by the given standard deviation (`random_std`).

        Args:
            shape (int or tuple of ints): The shape of the random array. If an integer is provided,
                it will be treated as a single-dimensional array.
            name (str or None): A name for the Container object. Default is None.
            random_std ( int or float): The standard deviation for generating random values.
                Default is 1.0.

        Returns:
            Container: A Container object containing the generated random array.

        Examples:
            >>> random_array = random((3, 4), name="random_data", random_std=2.0)
            >>> print(random_array)
            container([[ 0.47769941 -4.44432492 -1.20405922 -1.26902509]
                       [ 1.66314918 -2.66073428 -1.08072826  2.50640459]
                       [ 0.00681053  0.78007514 -2.9132976  -0.19908861]])

    """

    if isinstance(shape, int) and isinstance(name, int):
        warnings.warn("Your input is shape={} and name={}. "
                      "If you want to set the shape being ({}, {}), please use shape=({}, {})."
                      .format(shape, name, shape, name, shape, name), SyntaxWarning, stacklevel=3)

    if not isinstance(shape, (tuple, int)):
        raise ValueError("shape is expected tuple or int, but got {}".format(type(shape)))
    elif isinstance(shape, int):
        shape = (shape,)
    elif not all(isinstance(x, int) for x in shape):
        raise ValueError("shape should only int or tuple of ints, but got tuple having not int value.")

    if not isinstance(random_std, (float, int)):
        raise ValueError("random_std is expected int or float, but got {}".format(type(random_std)))
    elif isinstance(random_std, int):
        random_std = float(random_std)

    array_data = random_std * np.random.randn(*shape)

    return Container(array_data, name)


def random_int(low, high=None, size=None, name=None, dtype=None):
    """Generate a Container object containing random integers within the specified range.

        This function generates random integers within the specified range and wraps them in a Container object.

        Args:
            low (int): The lower bound (inclusive) of the generated integers.
            high (int or None): The upper bound (exclusive) of the generated integers.
                If not specified, `low` will be treated as the upper bound.
            size (int or tuple or None): The shape of the generated integers. It can be an integer or a tuple
                of integers. Default is None. Grant ``None``, it converts to (1,).
            name (str or None): The name of the Container object. Default is None.
            dtype (str or numpy.dtype or cupy.dtype or None): The data type of the generated integers.
                If not specified, the default data type will be used.

        Returns:
            Container: A Container object containing the generated random integers.

        Examples:
            >>> random_int_values = random_int(1, 10, size=(3, 2))
            >>> print(random_int_values)
            container([[6 8]
                       [8 8]
                       [4 3]])
    """

    if not isinstance(low, int):
        raise ValueError("low expects `int`, but got {}.".format(type(low)))
    if not isinstance(high, int):
        raise ValueError("high expects `int`, but got {}.".format(type(high)))

    if size is None:
        size = (1,)
    elif not isinstance(size, (int, tuple)):
        raise ValueError("size expects `int` or `tuple`, but got {}.".format(type(size)))
    elif isinstance(size, int):
        size = (size,)
    elif not all(isinstance(x, int) for x in size):
        raise ValueError("size can be tuple of ints, but input tuple has other than int.")

    if high is None:
        high = low
        low = 1

    if dtype is None:
        array_data = np.random.randint(low, high, size)
    else:
        array_data = np.random.randint(low, high, size, dtype)

    return Container(array_data, name=name)


def random_gen(*size, name=None):
    """Generate random values from a uniform distribution and wrap them in a Container object.

        This function generates random values from a uniform distribution and returns them as a Container object.

        Args:
            *size (int): The shape of the generated random values. It can be one or more integers.
            name (str or None): The name of the Container object. Default is None.

        Returns:
            Container: A Container object containing the generated random values.

        Example:
            >>> random_values = random_gen(3, 2)
            >>> print(random_values)
            container([[0.2100942  0.72216954]
                       [0.76960643 0.61244534]
                       [0.48039244 0.50885063]])
    """

    if not all(isinstance(x, int) for x in size):
        raise ValueError("size expected tuple of ints, but got other types")

    array_data = np.random.rand(*size)

    return Container(array_data, name=name)




from marquetry import cuda_backend
from marquetry.optimizer import Optimizer


class AdaGrad(Optimizer):
    """Adaptive Gradient Algorithm(AdaGrad) optimizer for updating model parameters during training.

        AdaGrad is an optimization algorithm that adapts the learning rate for each parameter
        individually based on historical gradient information.
        It is well-suited for problems with sparse gradients.
        The learning rate is adjusted based on the historical gradient information.

        Args:
            learning_rate (float): The learning rate for updating parameters.
                Default is 0.001.
            eps (float): A small value (epsilon) added to the denominator to prevent division by zero.
                Default is 1e-8.

        Tip:
            When you would like to optimize your model's parameter,
            please set the model to this using ``prepare`` method.

        Examples:
            >>> optimizer = AdaGrad()
            >>> model = marquetry.models.MLP([128, 256, 64, 10])
            >>> optimizer.prepare(model)
            >>> optimizer.update()

    """

    def __init__(self, learning_rate=0.001, eps=1e-8):
        super().__init__()
        self.lr = learning_rate
        self.eps = eps

        self.histories = {}

    def _update_one(self, param):
        h_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if h_key not in self.histories:
            self.histories[h_key] = xp.zeros_like(param.data)

        history = self.histories[h_key]
        grad = param.grad.data

        history += grad ** 2
        param.data -= self.lr * grad / (xp.sqrt(history) + self.eps)



from marquetry.datasets import csv_loader
from marquetry.datasets import fashion_mnist
from marquetry.datasets import mnist
from marquetry.datasets import trigonometric_curve
from marquetry.datasets import spiral
from marquetry.datasets import titanic

from marquetry.datasets.csv_loader import CsvLoader
from marquetry.datasets.fashion_mnist import FashionMNIST
from marquetry.datasets.mnist import MNIST
from marquetry.datasets.trigonometric_curve import TrigonometricCurve
from marquetry.datasets.spiral import Spiral
from marquetry.datasets.titanic import Titanic



import contextlib
import os
import sys

from marquetry import cuda_backend


class Config(object):
    """Configuration of the marquetry.

        This class implements the configuration of the marquetry dynamic computation graph.
        When a user use this class if user want to adjust themselves use case.

        Args:
            enable_backprop (bool): If this is ``True``, marquetry creates computation graph.
                Otherwise, never create computation graph.
            train (bool): If this is ``True``, marquetry's function computes it as train.
                Otherwise, computes it as predict mode.
            CACHE_DIR (str): The cache directory of the marquetry.
                The default path is ``~/.marquetry``.
    """

    enable_backprop = True
    train = True
    CUDA_ENABLE = cuda_backend.GPU_ENABLE
    CACHE_DIR = os.path.join(os.path.expanduser("~"), ".marquetry")

    def show(self, file=sys.stdout):
        keys = sorted(self.__dict__)
        _print_attrs(self, keys, file)


def _print_attrs(obj, keys, file):
    max_len = max(len(key) for key in keys)
    for key in keys:
        spacer = " " * (max_len - len(key))
        print("{}:{}{}".format(key, spacer, getattr(obj, key)), file=file)


config = Config()


@contextlib.contextmanager
def using_config(name, value, config_obj=config):
    """Context manager for temporarily modifying a configuration setting.

        This context manager allows you to temporarily change the value of a configuration setting
        within a specific context. You can use it to override certain configuration parameters for
        a specific block of code while ensuring that the original setting is restored after the block execution.

    Args:
        name (str): The name of the configuration setting to modify.
        value: The new value to set for the specified configuration setting.
        config_obj: The configuration object to modify.
            Defaults to the global configuration object defined in the ``config`` module.

    Yields:
        None

    Examples:
        >>> print(config.train)
        True
        >>> with using_config("train", False):
        >>>     print(config.train)
        False
        >>> print(config.train)
        True

    Note:
        This context manager is typically used in conjunction with
        the global configuration object defined in the 'config' module.
    """

    if hasattr(config_obj, name):
        old_value = getattr(config_obj, name)
        setattr(config_obj, name, value)
        try:
            yield
        finally:
            setattr(config_obj, name, old_value)
    else:
        setattr(config_obj, name, value)
        try:
            yield
        finally:
            delattr(config_obj, name)



from marquetry import Function


class Identity(Function):
    """Identity function.

        The Identity class simply returns its input unchanged. It is often used as a building block
        in neural networks to maintain the same input shape and value.
    """

    def forward(self, x):
        y = x.copy()

        self.retain_inputs(())
        return y

    def backward(self, input_data, grad_y):
        return grad_y


def identity(x):
    """Apply the Identity function to an input.

        The Identity function takes an input and returns it unchanged. It is often used as a
        simple way to pass data through a function without any modification.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input data to be passed through the Identity function.

        Returns:
            Container: A new Variable object containing the input data.

        Examples:
            >>> import marquetry as mq
            >>> input_data = mq.array([1, 2, 3])
            >>> result = mq.functions.identity(input_data)
            >>> print(result)
            container([1, 2, 3])
    """
    return Identity()(x)



import marquetry
from marquetry import functions
from marquetry import Layer
from marquetry.layers import Linear


class RNN(Layer):
    """Recurrent Neural Network (RNN) layer for sequence modeling.

        The RNN layer is a fundamental building block for processing sequential data.
        It performs a recurrent　computation on input data, allowing it to maintain hidden states
        and capture dependencies across time steps.

        Args:
            hidden_size (int): The size of the hidden state in the RNN layer.
            in_size (int or None): The size of the input data.

        Caution:
            in_size:
                This is automatically determined from the input data shape
                and does not need to be specified except a special use case.

        Attributes:
            h (marquetry.Container or None): The current hidden state of the RNN.

        Examples:
            >>> dataset = marquetry.datasets.SinCurve()
            >>> dataloader = marquetry.dataloaders.SeqDataLoader(dataset, batch_size=32)
            >>> model = (RNN(128), marquetry.layers.Linear(1))
            >>> loss = 0
            >>> for x, t in dataloader:
            >>>     for layer in model:
            >>>         x = layer(x)
            >>>     loss += functions.mean_squared_error(x, t)
            >>> loss
            container(261.4472488881584)

    """

    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.x2h = Linear(hidden_size, in_size=in_size)
        self.h2h = Linear(hidden_size, in_size=in_size, nobias=True)
        self.h = None

    def reset_state(self):
        """Reset the hidden state of the RNN."""
        self.h = None

    def set_state(self, h):
        """Set the hidden state of the RNN to the given value.

            Args:
                h (marquetry.Container):
                    The new hidden state.

        """

        if not isinstance(h, marquetry.Container):
            raise TypeError("hidden state type should be marquetry.Container, but got {}."
                            .format(type(h).__name__))

        self.h = h

    def forward(self, x):
        if self.h is None:
            new_hidden_state = functions.tanh(self.x2h(x))
        else:
            new_hidden_state = functions.tanh(self.x2h(x) + self.h2h(self.h))

        self.h = new_hidden_state

        return new_hidden_state



import marquetry
from marquetry import functions
from marquetry import Layer
from marquetry.layers import Linear


class GRU(Layer):
    """Gated Recurrent Unit (GRU) layer for sequence modeling.

        The GRU layer is a type of recurrent neural network (RNN) layer used for modeling sequential data. It is a variation
        of the Long Short-Term Memory (LSTM) layer and is capable of learning long-term dependencies in sequential data.

        Args:
            hidden_size (int): The size of the hidden state.
            in_size (int or None): The size of the input data.

        Caution:
            in_size:
                This is automatically determined from the input data shape
                and does not need to be specified except a special use case.

        Attributes:
            h (:class:`marquetry.Container`): The current hidden state.

        Examples:
            >>> dataset = marquetry.datasets.SinCurve()
            >>> dataloader = marquetry.dataloaders.SeqDataLoader(dataset, batch_size=32)
            >>> model = (GRU(128), marquetry.layers.Linear(1))
            >>> loss = 0
            >>> for x, t in dataloader:
            >>>     for layer in model:
            >>>         x = layer(x)
            >>>     loss += functions.mean_squared_error(x, t)
            >>> loss
            container(29.910813823186242)

    """

    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.hidden_size = hidden_size

        self.x2h = Linear(hidden_size, in_size=in_size)
        self.x2r = Linear(hidden_size, in_size=in_size)
        self.x2u = Linear(hidden_size, in_size=in_size)

        self.h2h = Linear(hidden_size, in_size=hidden_size, nobias=True)
        self.h2r = Linear(hidden_size, in_size=hidden_size, nobias=True)
        self.h2u = Linear(hidden_size, in_size=hidden_size, nobias=True)

        self.h = None

    def reset_state(self):
        """Reset the hidden state."""
        self.h = None

    def set_state(self, h: marquetry.Container):
        """Set the hidden state to a custom value.

            Args:
                h (marquetry.Container): The custom hidden state.
        """
        if not isinstance(h, marquetry.Container):
            raise ValueError("custom hidden state expected marquetry.Container type, but got {}"
                             .format(type(h).__name__))
        self.h = h

    def forward(self, x):
        if self.h is None:
            new_h = functions.tanh(self.x2h(x))

        else:
            reset_gate = functions.sigmoid(self.x2r(x) + self.h2r(self.h))
            new_h = functions.tanh(self.x2h(x) + self.h2h(reset_gate * self.h))
            update_gate = functions.sigmoid(self.x2u(x) + self.h2u(self.h))

            new_h = (1 - update_gate) * new_h + update_gate * self.h

        self.h = new_h

        return new_h



from marquetry.transformers.astypes import AsType
from marquetry.transformers.astypes import ToInt
from marquetry.transformers.astypes import ToFloat

from marquetry.transformers.compose import Compose

from marquetry.transformers.flatten import Flatten

from marquetry.transformers.normalize import Normalize



import setuptools
from setuptools import setup

long_description = """# Marquetry
Marquetry means **Yosegi-zaiku**, a traditional Japanese woodworking technique, in Japan.
It is a beautiful culture craft originated in Japan, which is a box or ornament or so by small wooden pieces.
The design is UNIQUE, it depends on the arrangement of the wood tips.
I believe Deep Learning is similar with the concept.
Deep Learning models are constructed through the combination of the layers or functions.
Just as a slight variation in arrangement can result in a completely distinct model.
I want you can enjoy the deep/machine learning journey like
you craft a **Marquetry** from combination of various wood tips.

## About this Framework
You can use this framework for help your learning **Machine/Deep Learning**.
This framework is written only **Python**, so you can understand the implementation easily if you are python engineer.
For simplify the construct, there are un-efficiency implementation.
I develop this framework to enjoy learning the construction of the machine/machine learning not **Practical Usage**.
I hope to enjoy your journey!"""

setup(
    name="marquetry",
    version="0.1.0",
    license="MIT",
    install_requires=[
        "numpy>=1.20.0",
        "pandas>=1.5.0",
        "Pillow>=9.2.0"
    ],
    description="Simple Machine Learning Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="SHIMA",
    maintainer="Little Tabby",
    author_email="shima@little-tabby.com",
    maintainer_email="engineer@little-tabby.com",
    url="https://github.com/little-tabby/Marquetry",
    download_url="https://github.com/little-tabby/Marquetry",
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    keywords="deeplearning ml neuralnetwork",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import typing

from marquetry import cuda_backend
from marquetry import Function


class ClassificationError(Function):
    """Compute classification error based on true labels and predictions.

        This class defines that computes the classification error based on the true labels and predicted labels.
        It supports an optional `ignore_label` parameter to exclude certain labels from the error calculation.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, ignore_label, dtype="float64"):
        self.ignore_label = ignore_label
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        self.retain_inputs(())
        if self.ignore_label is not None:
            mask = xp.asarray(t == self.ignore_label).astype("f")
            ignore_cnt = mask.sum()

            pred = xp.where(mask, self.ignore_label, y.argmax(axis=1).reshape(t.shape))

            count = t.size - xp.asarray(pred == t).sum()
            total = t.size - ignore_cnt

            if total == 0:
                return xp.asarray(0.0, dtype=self.dtype)
            else:
                return xp.asarray(float(count) / total, dtype=self.dtype)

        else:
            if y.size != t.size:
                pred = y.argmax(axis=1).reshape(t.shape)
            else:
                pred = y

            return xp.asarray((pred != t)).mean(dtype=self.dtype)


def classification_error(y, t, ignore_label: typing.Optional[int] = None, dtype="float64"):
    """Compute classification error based on true labels and predictions.


        This function defines that computes the classification error based on the true labels and predicted labels.
        It supports an optional `ignore_label` parameter to exclude certain labels from the error calculation.

        ClassificationError is calculated as:
            :math:`ClassificationError = (FP + FN) / (TP + FP + TN + FN)`

            (TP -> TruePositive, FP -> FalsePositive, TN -> TrueNegative, FN -> FalseNegative)


        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted labels.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.
            ignore_label (int or None):
                The label value to be ignored when calculating the classification error.
                If None, no labels are ignored.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The classification error based on the true
                labels and predicted labels, excluding labels specified by 'ignore_label' (if any).
        """

    return ClassificationError(ignore_label, dtype)(y, t)


class BinaryClassificationError(Function):
    """Compute binary classification error based on true labels and predicted values.

        This class defines a function that computes the binary classification error based on the true binary labels and
        predicted values. The binary classification error is calculated based on a specified threshold.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, threshold, dtype="float64"):
        if not 0. <= threshold <= 1.:
            raise ValueError("threshold should be in (0.0, 1.0), but got {}".format(threshold))

        self.threshold = threshold
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) <= 2

        pred = xp.asarray((y >= self.threshold), dtype="f")

        self.retain_inputs(())
        return xp.asarray((pred != t)).mean(dtype=self.dtype)


def binary_classification_error(y, t, threshold: float = 0.7, dtype="float64"):
    """Compute binary classification error based on true labels and predicted values.


        This function defines that computes the binary classification error based on the true binary labels and
        predicted values. The binary classification error is calculated based on a specified threshold.

        ClassificationError is calculated as:
            :math:`ClassificationError = (FP + FN) / (TP + FP + TN + FN)`

            (TP -> TruePositive, FP -> FalsePositive, TN -> TrueNegative, FN -> FalseNegative)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.
            threshold (float): The threshold value used to determine binary predictions.
                Defaults to 0.7.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The binary classification error based on the true binary labels and
                predicted values using the specified threshold.
    """
    return BinaryClassificationError(threshold, dtype)(y, t)



def reshape_sum_backward(grad_y, x_shape, axis, keepdims):
    """Reshape gradient appropriately for sum's backward."""

    ndim = len(x_shape)
    tupled_axis = axis
    if axis is None:
        tupled_axis = None
    elif not isinstance(axis, tuple):
        tupled_axis = (axis,)

    if not (ndim == 0 or tupled_axis is None or keepdims):
        actual_axis = [a if a >= 0 else a + ndim for a in tupled_axis]
        shape = list(grad_y.shape)
        for a in sorted(actual_axis):
            shape.insert(a, 1)
    else:
        shape = grad_y.shape

    grad_y = grad_y.reshape(shape)
    return grad_y


def max_backward_shape(x, axis):
    """Calc the origin input data's shape for max(min) backward."""

    if axis is None:
        axis = range(x.ndim)
    elif isinstance(axis, int):
        axis = (axis,)
    else:
        axis = axis

    shape = [s if ax not in axis else 1 for ax, s in enumerate(x.shape)]

    return shape



from marquetry import functions


class Min(functions.Max):
    """Calculate the minimum along the specified axis.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        y = x.min(axis=self.axis, keepdims=self.keepdims)

        self.retain_outputs((0,))
        return y


def min(x, axis=None, keepdims=False):
    """Calculate the minimum along the specified axis.

            Args:
                x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The input tensor.
                axis (int or tuple of ints or None): The axis or axes along which to find the minimum.
                keepdims (bool): If True, the output has the same number of dimensions as the input,
                    otherwise size 1's dimension is reduced.

            Returns:
                :class:`marquetry.Container`: The minimum value along the specified axis.

            Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> min(x)
                container(1)
                >>> min(x, axis=1)
                container([1 2])
                >>> min(x, axis=1, keepdims=True)
                container([[1]
                           [2]])

        """

    return Min(axis, keepdims)(x)



import marquetry
from marquetry import cuda_backend


def im2col_array(img, kernel_size, stride, pad, to_matrix=True):
    """The util of im2col main process"""

    batch_size, channels, height, weight = img.shape
    kernel_height, kernel_width = marquetry.utils.pair(kernel_size)
    stride_height, stride_width = marquetry.utils.pair(stride)
    padding_height, padding_width = marquetry.utils.pair(pad)

    # calc the output_size(output_height, output_width)
    out_height = marquetry.utils.get_convolution_outsize(height, kernel_height, stride_height, padding_height)
    out_width = marquetry.utils.get_convolution_outsize(weight, kernel_width, stride_width, padding_width)

    xp = cuda_backend.get_array_module(img)
    # Padding addition
    img = xp.pad(img, (
        (0, 0), (0, 0),
        (padding_height, padding_height + stride_height - 1),
        (padding_width, padding_width + stride_width - 1)), mode="constant", constant_values=(0,))

    # output array which has 6-dims to handle the local area extraction.
    col = xp.ndarray((batch_size, channels, kernel_height, kernel_width, out_height, out_width), dtype=img.dtype)

    for height in range(kernel_height):
        # The limit of the height first point(out_height means the kernel route point num respect stride_height.)
        # output_height * stride_height means the height point relative the origin first point.
        # (origin_point + relative_points)
        height_lim = height + out_height * stride_height

        for width in range(kernel_width):
            # the same mechanism as height_limit
            width_lim = width + out_width * stride_width

            col[:, :, height, width, :, :] = img[:, :, height:height_lim:stride_height, width:width_lim:stride_width]

    if to_matrix:
        # To more easy use, the 6-dims array convert to matrix having
        # (batch_size * output_height * output_width, channels * kernel_height * kernel_width)
        # convolution can be interpreted as a special case of the linear transformation.
        # The kernel can be changed the matrix which is (channels * kernel_height * kernel_width, out_channels)
        col = col.transpose((0, 4, 5, 1, 2, 3)).reshape((batch_size * out_height * out_width, -1))

    return col



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions
from marquetry import utils
from marquetry.functions.connection.convolution_2d_grad_w import Conv2DGradW


class Convolution2D(Function):
    """Perform 2D convolution on an input array.

        This class defines a function that performs 2D convolution on an input array or container using the given
        convolutional kernel and optional bias.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, stride=1, pad=0):
        super().__init__()
        self.stride = utils.pair(stride)
        self.pad = utils.pair(pad)

    def forward(self, x, w, b):
        xp = cuda_backend.get_array_module(x)

        kernel_height, kernel_width = w.shape[2:]
        col = utils.im2col_array(x, (kernel_height, kernel_width), self.stride, self.pad, to_matrix=False)

        y = xp.tensordot(col, w, ((1, 2, 3), (1, 2, 3)))
        if b is not None:
            y += b

        y = xp.rollaxis(y, 3, 1)

        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_x = functions.deconvolution_2d(
            grad_y, w, b=None, stride=self.stride, pad=self.pad, out_size=(x.shape[2], x.shape[3]))

        grad_w = Conv2DGradW(self)(x, grad_y)

        grad_b = None

        if b is not None:
            grad_b = grad_y.sum(axis=(0, 2, 3))

        return grad_x, grad_w, grad_b


def convolution_2d(x, w, b=None, stride=1, pad=0):
    """Perform 2D convolution on an input array or container.


        This class defines a function that performs 2D convolution on an input array or container using the given
        convolutional kernel and optional bias.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array or container to be convolved.
            w (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The convolutional kernel.
            b (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray` or None):
                The bias term. If None, no bias is added.
            stride (int or tuple of ints): The stride for moving the convolutional kernel.
                If an int is provided, it is converted to a tuple of the same value for both dimensions.
                Defaults to 1.
            pad (int or tuple of ints): The padding added to the input array or container before convolution.
                If an int is provided, it is converted to a tuple of the same value for both dimensions.
                Defaults to 0.

        Returns:
            :class:`marquetry.Container`:
                The result of 2D convolution on the input array using the provided kernel and bias (if any).

        Examples:
            >>> x = np.arange(0, 147).reshape(1, 3, 7, 7)
            >>> w = np.ones((16, 3, 2, 2))
            >>> convolution_2d(x, w).shape
            (1, 16, 6, 6)
    """

    return Convolution2D(stride, pad)(x, w, b)



from marquetry import cuda_backend


def impurity_criterion(target, criterion=None, target_type="classification"):
    """Compute an impurity criterion for a given target container.

        This function computes an impurity criterion for a given target container
        based on the specified criteria and target type.

        Args:
            target (:class:`numpy.ndarray`, or :class:`cupy.ndarray`):
                The target array for which the impurity criterion should be computed.
            criterion (str or None): The impurity criterion to use.
                For classification, options are "gini" or "entropy."
                For regression, options are "rss" (Residual Sum of Squares) or "mae" (Mean Absolute Error).
                Default is None, which selects the appropriate criterion based on the target_type.
            target_type (str): The type of the target container.
                Options are "classification" or "regression" Default is "classification".

        Returns:
            float: The computed impurity criterion.
    """

    if target_type == "classification":
        criterion = criterion if criterion is not None else "gini"
        return _classification_impurity_criterion(target, criterion)
    elif target_type == "regression":
        criterion = criterion if criterion is not None else "rss"
        return _regression_impurity_criterion(target, criterion)
    else:
        raise ValueError("target_type: {} is not supported.".format(target))


def _classification_impurity_criterion(target, criterion):
    """Compute a classification impurity criterion for a given target container.

        This function computes a classification impurity criterion for a given target container
        based on the specified criterion.

        Args:
            target (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container for which the impurity criterion should be computed.
            criterion (str): The impurity criterion to use.
                Options are "gini" or "entropy".

        Returns:
            float: The computed classification impurity criterion.

    """

    xp = cuda_backend.get_array_module(target)
    classes = xp.unique(target)
    num_samples = len(target)

    if criterion.lower() == "gini":
        result = 1.
        for class_num in classes:
            # calc each class rate
            rate = float(len(target[target == class_num])) / num_samples
            result -= rate ** 2

    elif criterion.lower() == "entropy":
        result = 0.
        for class_num in classes:
            # calc each class rate
            rate = float(len(target[target == class_num])) / num_samples
            result -= rate * xp.log2(rate)
    else:
        raise ValueError("{} is not supported as criterion.".format(criterion))

    return result


def _regression_impurity_criterion(target, criterion):
    """Compute a regression impurity criterion for a given target container.

        This function computes a regression impurity criterion for a given target container
        based on the specified criterion.

        Args:
            target (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container for which the impurity criterion should be computed.
            criterion (str): The impurity criterion to use.
                Options are "rss" (Residual Sum of Squares) or "mae" (Mean Absolute Error).

        Returns:
            float: The computed regression impurity criterion.
    """

    xp = cuda_backend.get_array_module(target)
    estimate_target = target.mean()
    if criterion.lower() == "rss":
        result = xp.sum((target - estimate_target) ** 2)
    elif criterion.lower() == "mae":
        result = xp.mean(xp.abs(target - estimate_target))
    else:
        raise ValueError("{} is not supported as criterion.".format(criterion))

    return result



from marquetry import cuda_backend
from marquetry import Function


class Tanh(Function):
    """Hyperbolic Tangent function.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.tanh(x)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = grad_y[0] * (1 - y ** 2)

        return grad_x


def tanh(x):
    """Hyperbolic Tangent function.

        This function's result is obtained -1.0 ~ 1.0.

        :math:`f(x) = {(exp(x) - exp(-x)) / (exp(x) + exp(-x))}`

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                   [ 2., -3.],
                   [-2.,  1.]], dtype=float32)
            >>> tanh(x, axis=1)
            container([[-0.7615942  0.       ]
                       [ 0.9640276 -0.9950548]
                       [-0.9640276  0.7615942]])

    """

    return Tanh()(x)



from marquetry import functions


def classification_cross_entropy(x, t):
    """Calculate the Cross-Entropy loss for classification tasks based on the dimensionality of the input.

        This function calculates the Cross-Entropy loss for classification problems,
        depending on the dimensionality of the input data.

        If the input has a single dimension or a single output unit,
        it calculates the :class:`marquetry.functions.sigmoid_cross_entropy` loss.

        If the input has multiple dimensions or output units,
        it calculates the :class:`marquetry.functions.softmax_cross_entropy` loss.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container`, :class:`numpy.ndarray`, or :class:`cupy.ndarray`):
                The true labels or binary targets.

        Returns:
            :class:`marquetry.Container`: The Cross-Entropy loss calculated
                based on the dimensionality of the input data.
    """

    if x.ndim == 1 or x.shape[1] == 1:
        return functions.sigmoid_cross_entropy(x, t)
    else:
        return functions.softmax_cross_entropy(x, t)



import unittest

import numpy as np

import marquetry.functions as funcs
from marquetry.utils import array_equal, gradient_check


class TestSin(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.sin(x)

        res = y.data
        expected = np.sin(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.sin(x)

        res = y.data
        expected = np.sin(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.sin, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.sin, x))


class TestCos(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.cos(x)

        res = y.data
        expected = np.cos(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.cos(x)

        res = y.data
        expected = np.cos(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.cos, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.cos, x))


class TestTan(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.tan(x)

        res = y.data
        expected = np.tan(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.tan(x)

        res = y.data
        expected = np.tan(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.rand(10, 30)

        self.assertTrue(gradient_check(funcs.tan, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.tan, x))


class TestTanh(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.tanh(x)

        res = y.data
        expected = np.tanh(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.tanh(x)

        res = y.data
        expected = np.tanh(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(10, 20)
        self.assertTrue(gradient_check(funcs.tanh, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.tanh, x))


class TestExp(unittest.TestCase):

    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.exp(x)

        res = y.data
        expected = np.exp(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.exp(x)

        res = y.data
        expected = np.exp(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.exp, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.exp, x))


class TestLog(unittest.TestCase):
    def test_forward1(self):
        x = np.array([1, 2, 3])
        y = funcs.log(x)

        res = y.data
        expected = np.log(x)

        self.assertTrue(array_equal(res, expected))

    def test_forward2(self):
        x = np.array(4)
        y = funcs.log(x)

        res = y.data
        expected = np.log(x)

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.abs(np.random.randn(3, 4))
        x = np.clip(x, 1e-8, 0.999)

        self.assertTrue(gradient_check(funcs.log, x))

    def test_backward2(self):
        x = np.abs(np.random.rand(1))
        x = np.clip(x, 1e-8, 0.999)

        self.assertTrue(gradient_check(funcs.log, x))


class TestAbsolute(unittest.TestCase):
    def test_forward1(self):
        x = np.array([[1, -2, 0], [-1, 1, 2]])
        expected = np.array([[1, 2, 0], [1, 1, 2]])

        y = funcs.absolute(x)

        self.assertTrue(array_equal(y.data, expected))

    def test_forward2(self):
        x = np.array(-1)
        expected = np.array(1)

        y = funcs.absolute(x)

        self.assertTrue(array_equal(y.data, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.absolute, x))

    def test_backward2(self):
        x = np.random.randn(1)

        self.assertTrue(gradient_check(funcs.absolute, x))


class TestAverage(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[2, 4, 6], [1, 2, 3]])
        expected = np.array(3.)

        y = funcs.average(x)

        self.assertTrue(array_equal(y.data, expected))

    def test_forward2(self):
        x = np.array([[2, 4, 6], [1, 2, 3]])
        expected = np.array([[4.], [2.]])

        axis = 1
        keepdims = True

        y = funcs.average(x, axis=axis, keepdims=keepdims)

        self.assertTrue(array_equal(y.data, expected))

    def test_forward3(self):
        x = np.array(8)
        expected = np.array(8.)

        y = funcs.average(x)

        self.assertTrue(array_equal(y.data, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.average, x))

    def test_backward2(self):
        x = np.random.randn(3, 4)

        f = lambda x: funcs.average(x, axis=1)

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        x = np.random.randn(3, 4)

        f = lambda x: funcs.average(x, axis=0, keepdims=True)

        self.assertTrue(gradient_check(f, x))


class TestReciprocal(unittest.TestCase):
    def test_forward1(self):
        x = np.array([[2, 4, 6], [1, 2, 3]])
        expected = np.array([[.5, .25, 1/6], [1, .5, 1/3]])

        y = funcs.reciprocal(x, dtype=np.float64)

        self.assertTrue(array_equal(y.data, expected))

    def test_forward2(self):
        x = np.array(8)
        expected = np.array(0.125)

        y = funcs.reciprocal(x)

        self.assertTrue(array_equal(y.data, expected))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        f = lambda x: funcs.reciprocal(x, dtype=np.float64)

        self.assertTrue(gradient_check(f, x))



import numpy as np


class Normalize(object):
    """A data transformation class for normalizing data by subtracting mean and dividing by standard deviation.

        Examples:
            >>> mean = x.mean(axis=1)
            >>> std = x.std(axis=1)
            >>> dataset = marquetry.datasets.MNIST(transform=Normalize(mean=mean, std=std))

    """

    def __init__(self, mean=0, std=1):
        self.mean = mean
        self.std = std

    def __call__(self, array):
        mean, std = self.mean, self.std

        if not np.isscalar(mean):
            mshape = [1] * array.ndim
            mshape[0] = len(array) if len(mean) == 1 else len(mean)
            mean = np.array(mean, dtype=array.dtype).reshape(*mshape)

        if not np.isscalar(std):
            rshape = [1] * array.ndim
            rshape[0] = len(array) if len(std) == 1 else len(std)
            std = np.array(std, dtype=array.dtype).reshape(*rshape)

        return (array - mean) / std



from typing import Sequence

import numpy as np

from marquetry import cuda_backend
from marquetry import Container


def numerical_grad(func, x, *args, **kwargs):
    """Numerical Gradient calculator"""
    eps = 1e-4

    x = x.data if isinstance(x, Container) else x
    xp = cuda_backend.get_array_module(x)
    if xp is not np:
        np_x = cuda_backend.as_numpy(x)
    else:
        np_x = x

    grad = xp.zeros_like(x)

    iters = np.nditer(np_x, flags=["multi_index"], op_flags=["readwrite"])
    while not iters.finished:
        index = iters.multi_index
        tmp_val = x[index].copy()

        x[index] = tmp_val + eps
        y1 = func(x, *args, **kwargs)
        if isinstance(y1, Container):
            y1 = y1.data

        y1 = y1.copy()

        x[index] = tmp_val - eps
        y2 = func(x, *args, **kwargs)
        if isinstance(y2, Container):
            y2 = y2.data

        y2 = y2.copy()

        if isinstance(y1, list):
            diff = 0
            for i in range(len(y1)):
                diff += (y1[i] - y2[i]).sum()
        else:
            diff = (y1 - y2).sum()

        if isinstance(diff, Container):
            diff = diff.data

        grad[index] = diff / (2 * eps)

        x[index] = tmp_val
        iters.iternext()

    return grad



from marquetry import cuda_backend
from marquetry import Function


class ReLU(Function):
    """Rectifier Linear Unit.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.maximum(x, 0.0)

        return y

    def backward(self, x, grad_y):
        x, = x
        mask = x > 0
        grad_x = grad_y[0] * mask

        return grad_x


def relu(x):
    """Rectified Linear Unit function.

        if x >= 0, :math:`f(x) = x` || if x < 0, :math:`f(x) = 0`

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                   [ 2., -3.],
                   [-2.,  1.]], dtype=float32)
            >>> relu(x)
            container([[0. 0.]
                       [2. 0.]
                       [0. 1.]])

    """
    return ReLU()(x)



from marquetry import cuda_backend
from marquetry import Function


class MeanAbsoluteError(Function):
    """Calculate the Mean Absolute Error (MAE) between two sets of values as loss function.

        This class computes the MAE between two sets of values.
        The MAE is a measure of the average absolute difference between corresponding elements of the input arrays.
        MAE is used as loss function when you want to use a data having outlier.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.

    """
    def forward(self, x0, x1):
        xp = cuda_backend.get_array_module(x0)
        diff = x0 - x1
        y = xp.absolute(diff).sum() / diff.dtype.type(diff.size)

        return y

    def backward(self, inputs, grad_y):
        xp = cuda_backend.get_array_module(inputs[0])
        x0, x1 = inputs

        diff = x0 - x1
        sign_map = xp.asarray(diff >= 0, dtype="f")
        sign_map -= .5
        sign_map *= 2

        coefficient = grad_y[0] * grad_y[0].dtype.type(1. / diff.size)
        grad_x = coefficient * sign_map

        return grad_x, -grad_x


def mean_absolute_error(x0, x1):
    """
        Calculate the Mean Absolute Error (MAE) between two sets of values.

        This function computes the MAE between two sets of values.
        The MAE is a measure of the average absolute　difference between corresponding elements of the input arrays.
        MAE is used as loss function when you want to use a data having outlier.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first set of values.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second set of values.

        Returns:
            :class:`marquetry.Container`: The Mean Absolute Error (MAE).
    """

    return MeanAbsoluteError()(x0, x1)



from marquetry import cuda_backend
from marquetry import Function


class Repeat(Function):
    """Copy and expand the array to the specify number and axis.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, repeats, axis):
        if isinstance(repeats, int):
            self.repeats = (repeats,)
        elif isinstance(repeats, (tuple, list)) and all(isinstance(x, int) for x in repeats):
            self.repeats = tuple(repeats)
        else:
            raise TypeError("repeats must be int or tuple if ints, but got {}".format(repeats))

        if not all(x >= 0 for x in self.repeats):
            raise ValueError("all repeats elements must be larger than 0.")

        if axis is not None and not isinstance(axis, int) or isinstance(axis, bool):
            raise TypeError("axis must be int or None.")

        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        repeats = self.repeats

        y = xp.repeat(x, repeats, self.axis)
        return y

    def backward(self, x, grad_y):
        x, = x
        grad_y, = grad_y

        grad_x = RepeatGrad(self.repeats, self.axis, x.shape, x.dtype)(grad_y)
        return grad_x


class RepeatGrad(Function):
    def __init__(self, repeats, axis, in_shape, in_dtype):
        self.repeats = repeats
        self.axis = axis
        if axis is not None and axis < 0:
            self.axis += len(in_shape)

        self.in_shape = in_shape
        self.in_dtype = in_dtype

    def forward(self, grad_y):
        xp = cuda_backend.get_array_module(grad_y)

        repeats = self.repeats
        axis = self.axis
        shape = list(self.in_shape)
        dtype = self.in_dtype

        if len(grad_y) == 0:
            grad_x = xp.zeros(shape, dtype)
            return grad_x

        if len(repeats) == 1:
            repeats = int(repeats[0])
            if axis is None:
                grad_x = grad_y.reshape(-1, repeats).sum(axis=1).reshape(shape)
            else:
                shape[axis:axis + 1] = [-1, repeats]
                grad_x = grad_y.reshape(shape).sum(axis=axis + 1)

            return grad_x

        if axis is None:
            position = 0
            grad_x = xp.zeros(shape, dtype).reshape(-1)

            for i, repeat_num in enumerate(repeats):
                grad_x[i] = xp.sum(grad_y[position:position + repeat_num])
                position += repeat_num

            grad_x = grad_x.reshape(shape)

        else:
            position = 0
            grad_x = xp.zeros(shape, dtype)
            source = [slice(None)] * axis + [None]
            distance = [slice(None)] * axis + [None]

            for i, repeat_num in enumerate(repeats):
                source[-1] = slice(position, position + repeat_num)
                distance[-1] = slice(i, i + 1)
                grad_x[tuple(distance)] = grad_y[tuple(source)].sum(axis=axis, keepdims=True)
                position += repeat_num

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        grad_grad_x = repeat(grad_grad_y[0], self.repeats, self.axis)

        return grad_grad_x


def repeat(x, repeats, axis):
    """Repeat and expand an input array along a specified axis.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be repeated and expanded.
            repeats (int): The number of times to repeat the input along the specified axis.
            axis (int): The axis along which the input should be repeated and expanded.

        Returns:
            A :class:`marquetry.Container` object which is the result of repeating and
            expanding the input array along the specified axis.

        Examples:
            >>> x = np.arange(0, 3).reshape(1, 3)
            array([[0, 1, 2]])
            >>> repeat(x, repeats=3, axis=0)
            container([[0 1 2]
                       [0 1 2]
                       [0 1 2]])
    """
    return Repeat(repeats, axis)(x)



import unittest

import numpy as np

import marquetry.container
from marquetry import Container
from marquetry.utils import gradient_check, array_equal


class TestAdd(unittest.TestCase):

    def test_forward1(self):
        x0 = np.array([1, 2, 3])
        x1 = Container(np.array([1, 2, 3]))
        y = x0 + x1

        res = y.data
        expected = np.array([2, 4, 6])

        self.assertTrue(array_equal(res, expected))

    def test_datatype(self):
        x = Container(np.array(2.0))
        y = x ** 2

        self.assertFalse(np.isscalar(y))

    def test_backward1(self):
        x = Container(np.random.randn(3, 3))
        y = np.random.randn(3, 3)
        f = lambda x: x + y

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = Container(np.random.randn(3, 3))
        y = np.random.randn(3, 1)
        f = lambda x: x + y

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 1)

        self.assertTrue(gradient_check(marquetry.functions.add, x, y))


class TestSub(unittest.TestCase):
    def test_forward1(self):
        x0 = np.array([1, 2, 3])
        x1 = Container(np.array([1, 2, 3]))
        y = x0 - x1

        res = y.data
        expected = np.array([0, 0, 0])

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = Container(np.random.randn(3, 3))
        y = np.random.randn(3, 3)
        f = lambda x: x + y

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = Container(np.random.randn(3, 3))
        y = np.random.randn(3, 1)
        f = lambda x: x + y

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 1)

        self.assertTrue(gradient_check(marquetry.functions.sub, x, y))


class TestMul(unittest.TestCase):

    def test_forward1(self):
        x0 = np.array([1, 2, 3])
        x1 = Container(np.array([1, 2, 3]))

        y = x0 * x1
        res = y.data
        expected = np.array([1, 4, 9])

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 3)
        f = lambda x: x * y

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 1)
        f = lambda x: x * y

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        x = np.random.randn(3, 1)
        y = np.random.randn(3, 3)
        f = lambda x: x * y

        self.assertTrue(gradient_check(f, x))


class TestDiv(unittest.TestCase):

    def test_forward1(self):
        x0 = np.array([1, 2, 3])
        x1 = Container(x0)
        y = x0 / x1

        res = y.data
        expected = np.array([1, 1, 1])

        self.assertTrue(array_equal(res, expected))

    def test_backward1(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 3)
        f = lambda x: x / y

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        x = np.random.randn(3, 3)
        y = np.random.randn(3, 1)
        f = lambda x: x / y

        self.assertTrue(gradient_check(f, x))

    def test_backward3(self):
        x = np.random.randn(3, 1)
        y = np.random.randn(3, 3)
        f = lambda x: x / y

        self.assertTrue(gradient_check(f, x))



from marquetry import cuda_backend
from marquetry import Function


class RelativeSquaredError(Function):
    """Calculate the Relative Squared Error (RSE) between predicted values and true values.

        This class defines a function that calculates the Relative Squared Error (RSE),
        a measure of the relative difference between the squared differences of predicted
        and true values relative to the squared differences of true values from their mean.
        RSE quantifies the average squared difference between predicted and true values
        relative to the average squared deviation of true values from their mean.
        It is used to assess the accuracy of regression models.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, multi_output):
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        assert y.size == t.size

        xp = cuda_backend.get_array_module(y)

        if t.ndim == 1:
            t = t.reshape((-1, 1))

        y = y.reshape(t.shape)

        correct_average = xp.mean(t, axis=0)

        squared_error = xp.sum(xp.square(y - t), axis=0)
        squared_deviation = xp.sum(xp.square(t - correct_average), axis=0)

        relative_squared_error_value = xp.where(squared_deviation != 0,
                                                squared_error / squared_deviation,
                                                1.0)

        if self.multi_output == "uniform_average":
            return xp.asarray(relative_squared_error_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(relative_squared_error_value, dtype=y.dtype)


def relative_squared_error(y, t, multi_output="uniform_average"):
    """Calculate the Relative Squared Error (RSE) between predicted values and true values.

        This function defines that calculates the Relative Squared Error (RSE),
        a measure of the relative difference between the squared differences of predicted
        and true values relative to the squared differences of true values from their mean.
        RSE quantifies the average squared difference between predicted and true values
        relative to the average squared deviation of true values from their mean.
        It is used to assess the accuracy of regression models.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the RSE for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average RSE over all samples. This is the default option.

                "raw_values": Return the raw RSE values for each sample.

        Returns:
            :class:`marquetry.Container`: The RSE based on the predicted values
                and true values. Lower values indicate better accuracy.
    """

    return RelativeSquaredError(multi_output)(y, t)



from marquetry import cuda_backend
from marquetry.optimizer import Optimizer


class MomentumSGD(Optimizer):
    """Momentum Stochastic Gradient Descent (Momentum SGD) optimizer for updating model parameters during training.

        Momentum SGD is an optimization algorithm that uses a moving average of gradients (momentum) to adaptively
        adjust the learning rates for each parameter.
        It helps accelerate convergence and avoid locally optimal solution.

        This algorithm was inspired by momentum in physics.

        Args:
            learning_rate (float): The learning rate for updating parameters.
                Default is 0.01.
            decay (float): The momentum decay factor, controlling the influence of past gradients.
                Default is 0.9.

        Tip:
            When you would like to optimize your model's parameter,
            please set the model to this using ``prepare`` method.

        Examples:
            >>> optimizer = MomentumSGD()
            >>> model = marquetry.models.MLP([128, 256, 64, 10])
            >>> optimizer.prepare(model)
            >>> optimizer.update()

    """

    def __init__(self, learning_rate=0.01, decay=0.9):
        super().__init__()
        self.lr = learning_rate
        self.momentum = decay

        self.momentum_vector = {}

    def _update_one(self, param):
        v_key = id(param)

        if v_key not in self.momentum_vector:
            xp = cuda_backend.get_array_module(param.data)
            self.momentum_vector[v_key] = xp.zeros_like(param.data)

        pre_vector = self.momentum_vector[v_key]
        pre_vector *= self.momentum
        pre_vector -= (1 - self.momentum) * param.grad.data

        param.data += pre_vector




import typing

from marquetry import cuda_backend
from marquetry import Function


class Accuracy(Function):
    """Compute the accuracy of predicted labels compared to true labels.


        This class defines a function that computes the accuracy of predicted labels compared to true labels.
        It supports an optional `ignore_label` parameter to exclude certain labels from the accuracy calculation.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, ignore_label, dtype="float64"):
        self.ignore_label = ignore_label
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        self.retain_inputs(())
        if self.ignore_label is not None:
            mask = xp.asarray(t == self.ignore_label).astype("f")
            ignore_cnt = mask.sum()

            pred = xp.where(mask, self.ignore_label, y.argmax(axis=1).reshape(t.shape))

            count = xp.asarray(pred == t).sum() - ignore_cnt
            total = t.size - ignore_cnt

            if total == 0:
                return xp.asarray(0.0, dtype=self.dtype)
            else:
                return xp.asarray(float(count) / total, dtype=self.dtype)

        else:
            if y.size != t.size:
                pred = y.argmax(axis=1).reshape(t.shape)
            else:
                pred = y

            return xp.asarray((pred == t)).mean(dtype=self.dtype)


def accuracy(y, t, ignore_label: typing.Optional[int] = None, dtype="float64"):
    """Compute the accuracy of predicted labels compared to true labels.


        This function defines that computes the accuracy of predicted labels compared to true labels.
        It supports an optional `ignore_label` parameter to exclude certain labels from the accuracy calculation.


        Accuracy is calculated as:
            :math:`Accuracy = (TP + TN) / (TP + FP + TN + FN)`

            (TP -> TruePositive, FP -> FalsePositive, TN -> TrueNegative, FN -> FalseNegative)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted labels.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.
            ignore_label (int or None):
                The label value to be ignored when calculating accuracy.
                If None, no labels are ignored.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The accuracy of the predicted labels compared to the true labels,
                excluding labels specified by 'ignore_label' (if any).
    """
    return Accuracy(ignore_label, dtype)(y, t)


class BinaryAccuracy(Function):
    """Compute binary accuracy based on a specified threshold.


        This class defines that computes the binary accuracy of predicted binary values compared to true binary labels.
        The binary accuracy is calculated based on a specified threshold.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, threshold: float, dtype="float64"):
        if not 0. <= threshold <= 1.:
            raise ValueError("threshold should be in (0.0, 1.0), but got {}".format(threshold))
        self.threshold = threshold
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) <= 2

        pred = xp.asarray((y >= self.threshold), dtype=y.dtype).reshape(t.shape)

        self.retain_inputs(())
        return xp.asarray((pred == t)).mean(dtype=self.dtype)


def binary_accuracy(y, t, threshold: float = 0.7, dtype="float64"):
    """Compute binary accuracy based on a specified threshold.


        This function defines that computes the binary accuracy of predicted binary values compared to true binary labels.
        The binary accuracy is calculated based on a specified threshold.

        Accuracy is calculated as:
            :math:`Accuracy = (TP + TN) / (TP + FP + TN + FN)`

            (TP -> TruePositive, FP -> FalsePositive, TN -> TrueNegative, FN -> FalseNegative)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted binary values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.
            threshold (float): The threshold value used to determine binary predictions.
                Defaults to 0.7.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The binary accuracy of the predicted binary
                values compared to the true binary labels using the specified threshold.
    """

    return BinaryAccuracy(threshold, dtype)(y, t)



import pandas as pd

from marquetry import Preprocess


class ColumnNormalize(Preprocess):
    """Column Normalization for Specified Columns in a DataFrame.

        ColumnNormalize is a preprocessing step used to normalize specified columns in a Pandas DataFrame.
        It scales the values in the selected columns to have a mean of 0 and a standard deviation of 1.

        Args:
            norm_target_column (list): A list of column names to be normalized.
            name (str): The name of the ColumnNormalize instance.
            is_train (bool): A boolean flag indicating whether
                this ColumnNormalize instance is for training or inference.
            skip_nan (bool): If True, columns containing NaN (null) values will be skipped during normalization.
                If False, NaN values will raise an error.
                Defaults to True.

        Examples:
            >>> import pandas as pd
            >>> from marquetry.preprocesses import ColumnNormalize
            >>> data = pd.DataFrame({'FeatureA': [1.0, 2.0, 3.0, 4.0], 'FeatureB': [10.0, 20.0, 30.0, 40.0]})
            >>> normalizer = ColumnNormalize(norm_target_column=['FeatureA', 'FeatureB'], name='normalizer', is_train=True)
            >>> normalized_data = normalizer(data)
            >>> print(normalized_data)
               FeatureA  FeatureB
            0 -1.161895 -1.161895
            1 -0.387298 -0.387298
            2  0.387298  0.387298
            3  1.161895  1.161895

        Note:
            ColumnNormalize scales the values in the specified columns to have a mean of 0 and
            a standard deviation of 1.
            It provides an option to skip columns containing NaN (null) values during normalization.
    """

    _label = "pre_cn"

    def __init__(self, norm_target_column: list, name: str, is_train: bool, skip_nan: bool = True):
        super().__init__(name, is_train)
        self._norm_target_column = norm_target_column
        self._skip_nan = skip_nan

    def process(self, data: pd.DataFrame):
        """Process the input DataFrame by normalizing specified columns.

            Args:
                data (:class:`pandas.DataFrame`): The input DataFrame with columns to be normalized.

            Returns:
                pd.DataFrame: The DataFrame with specified columns normalized.

            Caution:
                Generally, process will be called by ``Marquetry`` core.
                Please call like the ``EXAMPLES``.

        """

        if len(self._norm_target_column) == 0:
            return data

        if self._statistic_data is None:
            self._calc_statistic(data)

        self._validate_values()
        normalized_data = data.copy()

        for column in self._norm_target_column:
            tmp_dict = self._statistic_data[column]

            normalized_data.loc[:, column] = (
                    (data.loc[:, column] - tmp_dict["average_value"]) / tmp_dict["standard_deviation"])

        return normalized_data

    def _validate_values(self):
        exist_statistic_columns = set([key for key, value in self._statistic_data.items() if value != {}])
        input_target_columns = set(self._norm_target_column)

        if exist_statistic_columns != input_target_columns:
            raise ValueError("saved statistic data's target columns is {} but you input {}. "
                             .format(exist_statistic_columns, input_target_columns) + self._msg)

        return

    def _calc_statistic(self, data: pd.DataFrame):
        normalize_dict = {}

        std_data = data.loc[:, tuple(self._norm_target_column)].std()
        mean_data = data.loc[:, tuple(self._norm_target_column)].mean()

        for column in list(data.columns):
            if column not in self._norm_target_column:
                normalize_dict[column] = {}
                continue

            if not self._skip_nan and data[column].isna().sum() != 0:
                raise ValueError("input data has null value, but it can't be skipped by user configure "
                                 "so the normalize can't be done expected.")

            tmp_std = std_data[column]
            tmp_mean = mean_data[column]

            tmp_dict = {
                "standard_deviation": tmp_std,
                "average_value": tmp_mean
            }

            normalize_dict[column] = tmp_dict

        self._save_statistic(normalize_dict)
        self._statistic_data = normalize_dict



from marquetry import Function


class LeakyReLU(Function):
    """Leaky Rectifier Linear Unit.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, slope):
        self.slope = slope

    def forward(self, x):
        y = x.copy()
        y[x <= 0] *= self.slope

        return y

    def backward(self, x, grad_y):
        mask = (x[0] > 0).astype(grad_y[0].dtype)
        mask[mask <= 0] = self.slope

        grad_x = grad_y[0] * mask

        return grad_x


def leaky_relu(x, slope=0.2):
    """Leaky Rectified Linear Unit function.

        This function is improved version of the ReLU(Rectified Linear Unit).

        if x >= 0, :math:`f(x) = x` || if x < 0, :math:`f(x) = slope * x`
            - The ``slope`` is a small constant value the default value is `0.2`.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.
            slope (float): small constant value. The default is 0.2.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                   [ 2., -3.],
                   [-2.,  1.]], dtype=float32)
            >>> leaky_relu(x, slope=0.2)
            container([[-0.2  0. ]
                       [ 2.  -0.6]
                       [-0.4  1. ]])

    """

    return LeakyReLU(slope)(x)
